/*
 * This file is a part of Winix
 * and is not publicly distributed
 * 
 * Copyright (c) 2012, Tomasz Sowa
 * All rights reserved.
 * 
 */
 
#ifndef headerfile_winix_templates_textextstream
#define headerfile_winix_templates_textextstream

#include <ctime>
#include "core/textstream.h"
#include "textstream/textstream.h"



/*
	TexTextStream is used as a buffer for creating a TeX input
	By default all operators<< escape its string arguments. If you don't want
	to escape an argument you should use a helper function R() (raw argument)
	note: you have to define the function yourself, we do not provide it
	because such a short name would make a mess in namespaces

	sample:
	create a helper function R as follows:

		template<class RawType>
		TexTextStream::RawText<RawType> R(const RawType & par)
		{
			return TexTextStream::RawText<RawType>(par);
		}
	
	now you can use TexTextStream in an easy way:
		TexTextStream page;
		std::string key = "some text with $# ^{} tex \\ special characters";
		page << key << R("\\def\\myfun{...}");

		everything in 'key' is property escaped for using with TeX
*/
class TexTextStream : public TextStream<std::wstring>
{
public:


	TexTextStream();


	/*
		a helper struct to select a proper operator<<
		(for non-escaping versions of these operators)
	*/
	template<class RawType>
	struct RawText
	{
		const RawType & par;

		RawText(const RawText<RawType> & p) : par(p.par) {}
		RawText(const RawType & p) : par(p) {}
	};



	/*
		without escaping
	*/
	TexTextStream & PutChar(char);
	TexTextStream & PutChar(wchar_t);

	TexTextStream & PutText(const char *);
	TexTextStream & PutText(const char *, size_t len);
	TexTextStream & PutText(const std::string *);
	TexTextStream & PutText(const std::string &);
	TexTextStream & PutText(const wchar_t * str);
	TexTextStream & PutText(const wchar_t * str, size_t len);
	TexTextStream & PutText(const std::wstring * str);
	TexTextStream & PutText(const std::wstring & str);

	/*
		we need this template operator for such calling:
		HtmlTextStream_object << R("some string");
		"some string" is actually a table (not a pointer)
	*/
	template<size_t str_size>
	TexTextStream & operator<<(const RawText<char [str_size]> & raw)	{ return PutText(raw.par); }

	template<size_t str_size>
	TexTextStream & operator<<(const RawText<wchar_t [str_size]> & raw)	{ return PutText(raw.par); }

	TexTextStream & operator<<(const RawText<const char*> & raw);
	TexTextStream & operator<<(const RawText<const wchar_t*> & raw);
	TexTextStream & operator<<(RawText<const std::string*> raw);
	TexTextStream & operator<<(RawText<const std::wstring*> raw);
	TexTextStream & operator<<(RawText<std::string> raw);
	TexTextStream & operator<<(RawText<std::wstring> raw);

	TexTextStream & operator<<(RawText<char> raw);
	TexTextStream & operator<<(RawText<wchar_t> raw);
	TexTextStream & operator<<(RawText<int> raw);
	TexTextStream & operator<<(RawText<long> raw);
	TexTextStream & operator<<(RawText<unsigned int> raw);
	TexTextStream & operator<<(RawText<unsigned long> raw);
	TexTextStream & operator<<(RawText<double> raw);
	TexTextStream & operator<<(RawText<void*> raw);

	template<typename arg_char_type, size_t arg_stack_size, size_t arg_heap_block_size>
	TexTextStream & operator<<(RawText<PT::TextStreamBase<arg_char_type, arg_stack_size, arg_heap_block_size> > raw);


	// 'write' don't escapes too
	// with these methods you can write a zero character too
	TexTextStream & Write(const char * buf, size_t len);
	TexTextStream & Write(const wchar_t * buf, size_t len);
	// for compatibility with standard library (Ezc uses it)
	TexTextStream & write(const char * buf, size_t len);
	TexTextStream & write(const wchar_t * buf, size_t len);


	/*
		with escaping
	*/

	TexTextStream & ETextPutChar(char c);
	TexTextStream & ETextPutChar(wchar_t c);

	TexTextStream & EPutText(const char * str);
	TexTextStream & EPutText(const char * str, size_t len);
	TexTextStream & EPutText(const std::string * str);
	TexTextStream & EPutText(const std::string & str);
	TexTextStream & EPutText(const wchar_t * str);
	TexTextStream & EPutText(const wchar_t * str, size_t len);
	TexTextStream & EPutText(const std::wstring * str);
	TexTextStream & EPutText(const std::wstring & str);


	TexTextStream & operator<<(const char * str);
	TexTextStream & operator<<(const std::string * str);
	TexTextStream & operator<<(const std::string & str);
	TexTextStream & operator<<(const wchar_t * str);
	TexTextStream & operator<<(const std::wstring * str);
	TexTextStream & operator<<(const std::wstring & str);
	TexTextStream & operator<<(char);
	TexTextStream & operator<<(wchar_t);
	TexTextStream & operator<<(int);
	TexTextStream & operator<<(long);
	TexTextStream & operator<<(unsigned int);
	TexTextStream & operator<<(unsigned long);
	TexTextStream & operator<<(double);
	TexTextStream & operator<<(const void *);
	TexTextStream & operator<<(const PT::Space & space);
	TexTextStream & operator<<(const PT::Date & Date);

	template<typename arg_char_type, size_t arg_stack_size, size_t arg_heap_block_size>
	TexTextStream & operator<<(const PT::TextStreamBase<arg_char_type, arg_stack_size, arg_heap_block_size> & arg);

private:

	TextStream<std::wstring> tmp_stream;

};


template<typename arg_char_type, size_t arg_stack_size, size_t arg_heap_block_size>
TexTextStream & TexTextStream::operator<<(RawText<PT::TextStreamBase<arg_char_type, arg_stack_size, arg_heap_block_size> > raw)
{
	TextStream<std::wstring>::operator<<(raw.par);

return *this;
}


template<typename arg_char_type, size_t arg_stack_size, size_t arg_heap_block_size>
TexTextStream & TexTextStream::operator<<(const PT::TextStreamBase<arg_char_type, arg_stack_size, arg_heap_block_size> & arg)
{
typename PT::TextStreamBase<arg_char_type, arg_stack_size, arg_heap_block_size>::const_iterator i;

	for(i=arg.begin() ; i != arg.end() ; ++i)
		ETextPutChar(*i);

return *this;
}



#endif

